import { getModifierState } from '../utils';
import { is_equal } from '../positioning';
import { Throttler } from '@lumino/polling';
import { CodeMirrorIntegration } from '../editor_integration/codemirror';
import { FeatureSettings } from '../feature';
import { EditorTooltipManager } from '../components/free_tooltip';
import { IRenderMimeRegistry } from '@jupyterlab/rendermime';
import { ILSPFeatureManager, PLUGIN_ID } from '../tokens';
import { ISettingRegistry } from '@jupyterlab/settingregistry';
import { LabIcon } from '@jupyterlab/ui-components';
import hoverSvg from '../../style/icons/hover.svg';
import { PositionConverter } from '../converter';
export const hoverIcon = new LabIcon({
    name: 'lsp:hover',
    svgstr: hoverSvg
});
class ResponseCache {
    constructor(maxSize) {
        this.maxSize = maxSize;
        this._data = [];
    }
    get data() {
        return this._data;
    }
    store(item) {
        if (this._data.length >= this.maxSize) {
            this._data.shift();
        }
        this._data.push(item);
    }
    clean() {
        this._data = [];
    }
}
function to_markup(content) {
    if (typeof content === 'string') {
        // coerce to MarkedString  object
        return {
            kind: 'plaintext',
            value: content
        };
    }
    else {
        return {
            kind: 'markdown',
            value: '```' + content.language + '\n' + content.value + '\n```'
        };
    }
}
export class HoverCM extends CodeMirrorIntegration {
    constructor() {
        super(...arguments);
        this.onKeyDown = (event) => {
            if (getModifierState(event, this.modifierKey) &&
                typeof this.last_hover_character !== 'undefined') {
                // does not need to be shown if it is already visible (otherwise we would be creating an identical tooltip again!)
                if (this.tooltip && this.tooltip.isVisible && !this.tooltip.isDisposed) {
                    return;
                }
                const document = this.virtual_editor.document_at_root_position(this.last_hover_character);
                let response_data = this.restore_from_cache(document, this.virtual_position);
                if (response_data == null) {
                    return;
                }
                event.stopPropagation();
                this.handleResponse(response_data, this.last_hover_character, true);
            }
        };
        this.onMouseLeave = (event) => {
            this.remove_range_highlight();
            this.maybeHideTooltip(event.relatedTarget);
        };
        this.on_hover = async () => {
            return await this.connection.getHoverTooltip(this.virtual_position, 
            // this might be wrong - should not it be using the specific virtual document?
            this.virtual_document.document_info, false);
        };
        /**
         * Underlines the word if a tooltip is available.
         * Displays tooltip if asked to do so.
         *
         * Returns true is the tooltip was shown.
         */
        this.handleResponse = (response_data, root_position, show_tooltip) => {
            let response = response_data.response;
            // testing for object equality because the response will likely be reused from cache
            if (this.last_hover_response != response) {
                this.remove_range_highlight();
                this.hover_marker = this.highlight_range(response_data.editor_range, 'cm-lsp-hover-available');
            }
            this.last_hover_response = response;
            if (show_tooltip) {
                this.lab_integration.tooltip.remove();
                const markup = HoverCM.get_markup_for_hover(response);
                let editor_position = this.virtual_editor.root_position_to_editor(root_position);
                this.tooltip = this.lab_integration.tooltip.create({
                    markup,
                    position: editor_position,
                    ce_editor: response_data.ce_editor,
                    adapter: this.adapter,
                    className: 'lsp-hover'
                });
                return true;
            }
            return false;
        };
        this.updateUnderlineAndTooltip = (event) => {
            try {
                return this._updateUnderlineAndTooltip(event);
            }
            catch (e) {
                if (!(e.message === 'Cell not found in cell_line_map' ||
                    e.message === "Cannot read property 'string' of undefined")) {
                    this.console.warn(e);
                }
            }
        };
        this.remove_range_highlight = () => {
            if (this.hover_marker) {
                this.hover_marker.clear();
                this.hover_marker = null;
                this.last_hover_response = undefined;
                this.last_hover_character = undefined;
            }
        };
    }
    get modifierKey() {
        return this.settings.composite.modifierKey;
    }
    get lab_integration() {
        return super.lab_integration;
    }
    get settings() {
        return super.settings;
    }
    restore_from_cache(document, virtual_position) {
        const { line, ch } = virtual_position;
        const matching_items = this.cache.data.filter(cache_item => {
            if (cache_item.document !== document) {
                return false;
            }
            let range = cache_item.response.range;
            return (line >= range.start.line &&
                line <= range.end.line &&
                (line != range.start.line || ch >= range.start.character) &&
                (line != range.end.line || ch <= range.end.character));
        });
        if (matching_items.length > 1) {
            this.console.warn('Potential hover cache malfunction: ', virtual_position, matching_items);
        }
        return matching_items.length != 0 ? matching_items[0] : null;
    }
    register() {
        this.cache = new ResponseCache(this.settings.composite.cacheSize);
        this.wrapper_handlers.set('mousemove', event => {
            // as CodeMirror.Editor does not support mouseleave nor mousemove,
            // we simulate the mouseleave for the editor in wrapper's mousemove;
            // this is used to hide the tooltip on leaving cells in notebook
            this.updateUnderlineAndTooltip(event)
                .then(keep_tooltip => {
                if (!keep_tooltip) {
                    this.maybeHideTooltip(event.target);
                }
            })
                .catch(this.console.warn);
        });
        this.wrapper_handlers.set('mouseleave', this.onMouseLeave);
        // show hover after pressing the modifier key
        // TODO: when the editor (notebook or file editor) is not focused, the keydown event is not getting to us
        //  (probably getting captured by lab); this gives subpar experience when using hover in two editors open
        //  side-by-side, BUT this does not happen for mousemove which properly reads keyModifier from the event
        //  (so this is no too bad as most of the time the user will get the desired outcome - they just need to
        //  budge the mice when holding ctrl if looking at a document which is not active).
        // whether the editor is focused
        this.wrapper_handlers.set('keydown', this.onKeyDown);
        // or just the wrapper (e.g. the notebook but no cell active)
        this.editor_handlers.set('keydown', (instance, event) => this.onKeyDown(event));
        this.debounced_get_hover = this.create_throttler();
        this.settings.changed.connect(() => {
            this.cache.maxSize = this.settings.composite.cacheSize;
            this.debounced_get_hover = this.create_throttler();
        });
        super.register();
    }
    maybeHideTooltip(mouse_target) {
        if (typeof this.tooltip !== 'undefined' &&
            mouse_target !== this.tooltip.node) {
            this.tooltip.dispose();
        }
    }
    create_throttler() {
        return new Throttler(this.on_hover, {
            limit: this.settings.composite.throttlerDelay,
            edge: 'trailing'
        });
    }
    afterChange(change, root_position) {
        super.afterChange(change, root_position);
        // reset cache on any change in the document
        this.cache.clean();
        this.last_hover_character = undefined;
        this.remove_range_highlight();
    }
    static get_markup_for_hover(response) {
        let contents = response.contents;
        // this causes the webpack to fail "Module not found: Error: Can't resolve 'net'" for some reason
        // if (lsProtocol.MarkedString.is(contents))
        ///  contents = [contents];
        if (typeof contents === 'string') {
            contents = [contents];
        }
        if (!Array.isArray(contents)) {
            return contents;
        }
        let markups = contents.map(to_markup);
        if (markups.every(markup => markup.kind == 'plaintext')) {
            return {
                kind: 'plaintext',
                value: markups.map(markup => markup.value).join('\n')
            };
        }
        else {
            return {
                kind: 'markdown',
                value: markups.map(markup => markup.value).join('\n\n')
            };
        }
    }
    is_token_empty(token) {
        return token.string.length === 0;
        // TODO  || token.type.length === 0? (sometimes the underline is shown on meaningless tokens)
    }
    is_event_inside_visible(event) {
        let target = event.target;
        return target.closest('.CodeMirror-sizer') != null;
    }
    is_useful_response(response) {
        return (response &&
            response.contents &&
            !(Array.isArray(response.contents) && response.contents.length === 0));
    }
    /**
     * Returns true if the tooltip should stay.
     */
    async _updateUnderlineAndTooltip(event) {
        const target = event.target;
        // if over an empty space in a line (and not over a token) then not worth checking
        if (target.classList.contains('CodeMirror-line')) {
            this.remove_range_highlight();
            return;
        }
        const show_tooltip = getModifierState(event, this.modifierKey);
        // currently the events are coming from notebook panel; ideally these would be connected to individual cells,
        // (only cells with code) instead, but this is more complex to implement right. In any case filtering
        // is needed to determine in hovered character belongs to this virtual document
        let root_position = this.position_from_mouse(event);
        // happens because mousemove is attached to panel, not individual code cells,
        // and because some regions of the editor (between lines) have no characters
        if (root_position == null) {
            this.remove_range_highlight();
            return;
        }
        let token = this.virtual_editor.getTokenAt(root_position);
        let document = this.virtual_editor.document_at_root_position(root_position);
        if (this.is_token_empty(token) ||
            document !== this.virtual_document ||
            !this.is_event_inside_visible(event)) {
            this.remove_range_highlight();
            return;
        }
        if (!is_equal(root_position, this.last_hover_character)) {
            let virtual_position = this.virtual_editor.root_position_to_virtual_position(root_position);
            this.virtual_position = virtual_position;
            this.last_hover_character = root_position;
            let response_data = this.restore_from_cache(document, virtual_position);
            if (response_data == null) {
                let response = await this.debounced_get_hover.invoke();
                if (this.is_useful_response(response)) {
                    let ce_editor = this.virtual_editor.get_editor_at_root_position(root_position);
                    let cm_editor = this.virtual_editor.ce_editor_to_cm_editor.get(ce_editor);
                    let editor_range = this.get_editor_range(response, root_position, token, cm_editor);
                    response_data = {
                        response: this.add_range_if_needed(response, editor_range, ce_editor),
                        document: document,
                        editor_range: editor_range,
                        ce_editor: ce_editor
                    };
                    this.cache.store(response_data);
                }
                else {
                    this.remove_range_highlight();
                    return;
                }
            }
            return this.handleResponse(response_data, root_position, show_tooltip);
        }
        else {
            return true;
        }
    }
    remove() {
        this.cache.clean();
        this.remove_range_highlight();
        this.debounced_get_hover.dispose();
        super.remove();
        // just to be sure
        this.debounced_get_hover = null;
        this.remove_range_highlight = null;
        this.handleResponse = null;
        this.on_hover = null;
    }
    get_editor_range(response, position, token, cm_editor) {
        if (typeof response.range !== 'undefined') {
            return this.range_to_editor_range(response.range, cm_editor);
        }
        // construct the range manually using the token information
        let start_in_root = {
            line: position.line,
            ch: token.start
        };
        let end_in_root = {
            line: position.line,
            ch: token.end
        };
        return {
            start: this.virtual_editor.root_position_to_editor(start_in_root),
            end: this.virtual_editor.root_position_to_editor(end_in_root),
            editor: cm_editor
        };
    }
    add_range_if_needed(response, editor_range, ce_editor) {
        if (typeof response.range !== 'undefined') {
            return response;
        }
        return Object.assign(Object.assign({}, response), { range: {
                start: PositionConverter.cm_to_lsp(this.virtual_editor.root_position_to_virtual_position(this.virtual_editor.transform_from_editor_to_root(ce_editor, editor_range.start))),
                end: PositionConverter.cm_to_lsp(this.virtual_editor.root_position_to_virtual_position(this.virtual_editor.transform_from_editor_to_root(ce_editor, editor_range.end)))
            } });
    }
}
class HoverLabIntegration {
    constructor(app, settings, renderMimeRegistry) {
        this.tooltip = new EditorTooltipManager(renderMimeRegistry);
    }
}
const FEATURE_ID = PLUGIN_ID + ':hover';
export const HOVER_PLUGIN = {
    id: FEATURE_ID,
    requires: [ILSPFeatureManager, ISettingRegistry, IRenderMimeRegistry],
    autoStart: true,
    activate: (app, featureManager, settingRegistry, renderMimeRegistry) => {
        const settings = new FeatureSettings(settingRegistry, FEATURE_ID);
        const labIntegration = new HoverLabIntegration(app, settings, renderMimeRegistry);
        featureManager.register({
            feature: {
                editorIntegrationFactory: new Map([['CodeMirrorEditor', HoverCM]]),
                id: FEATURE_ID,
                name: 'LSP Hover tooltip',
                labIntegration: labIntegration,
                settings: settings
            }
        });
    }
};
//# sourceMappingURL=hover.js.map